import datetime
import hashlib
import logging

from ceilometerclient import client as ceilometer_client
from ceilometerclient import exc as ceilometer_exceptions
from cinderclient import exceptions as cinder_exceptions
from cinderclient.v2 import client as cinder_client
from django.core.cache import cache
from django.db import transaction
from django.utils import timezone
from glanceclient import exc as glance_exceptions
from glanceclient.v2 import client as glance_client
from keystoneauth1 import session as keystone_session
from keystoneauth1.identity import v3
from keystoneclient import exceptions as keystone_exceptions
from keystoneclient.v3 import client as keystone_client
from neutronclient.client import exceptions as neutron_exceptions
from neutronclient.v2_0 import client as neutron_client
from novaclient import client as nova_client
from novaclient import exceptions as nova_exceptions
from requests import ConnectionError
import six

from waldur_core.structure import ServiceBackend
from waldur_core.structure.exceptions import SerializableBackendError
from waldur_openstack.openstack.models import Tenant

logger = logging.getLogger(__name__)


class OpenStackBackendError(SerializableBackendError):
    pass


class OpenStackSessionExpired(OpenStackBackendError):
    pass


class OpenStackAuthorizationFailed(OpenStackBackendError):
    pass


class OpenStackSession(dict):
    """ Serializable session """

    def __init__(self, ks_session=None, verify_ssl=False, **credentials):
        self.keystone_session = ks_session
        if not self.keystone_session:
            auth_plugin = v3.Password(**credentials)
            self.keystone_session = keystone_session.Session(auth=auth_plugin, verify=verify_ssl)

        try:
            # This will eagerly sign in throwing AuthorizationFailure on bad credentials
            self.keystone_session.get_auth_headers()
        except keystone_exceptions.ClientException as e:
            six.reraise(OpenStackAuthorizationFailed, e)

        for opt in ('auth_ref', 'auth_url', 'project_id', 'project_name', 'project_domain_name'):
            self[opt] = getattr(self.auth, opt)

    def __getattr__(self, name):
        return getattr(self.keystone_session, name)

    @classmethod
    def recover(cls, session, verify_ssl=False):
        if not isinstance(session, dict) or not session.get('auth_ref'):
            raise OpenStackBackendError('Invalid OpenStack session')

        args = {
            'auth_url': session['auth_url'],
            'token': session['auth_ref'].auth_token,
        }
        if session.get('project_id'):
            args['project_id'] = session['project_id']
        elif session.get('project_name') and session.get('project_domain_name'):
            args['project_name'] = session['project_name']
            args['project_domain_name'] = session['project_domain_name']

        ks_session = keystone_session.Session(auth=v3.Token(**args), verify=verify_ssl)
        return cls(ks_session=ks_session)

    def validate(self):
        if self.auth.auth_ref.expires > timezone.now() + datetime.timedelta(minutes=10):
            return True

        raise OpenStackSessionExpired('OpenStack session is expired')

    def __str__(self):
        return str({k: v if k != 'password' else '***' for k, v in self.items()})


class OpenStackClient(object):
    """ Generic OpenStack client. """

    def __init__(self, session=None, verify_ssl=False, **credentials):
        self.verify_ssl = verify_ssl
        if session:
            if isinstance(session, dict):
                logger.debug('Trying to recover OpenStack session.')
                self.session = OpenStackSession.recover(session, verify_ssl=verify_ssl)
                self.session.validate()
            else:
                self.session = session
        else:
            try:
                self.session = OpenStackSession(verify_ssl=verify_ssl, **credentials)
            except AttributeError as e:
                logger.error('Failed to create OpenStack session.')
                six.reraise(OpenStackBackendError, e)

    @property
    def keystone(self):
        return keystone_client.Client(session=self.session.keystone_session, interface='public')

    @property
    def nova(self):
        try:
            return nova_client.Client(version='2', session=self.session.keystone_session, endpoint_type='publicURL')
        except nova_exceptions.ClientException as e:
            logger.exception('Failed to create nova client: %s', e)
            six.reraise(OpenStackBackendError, e)

    @property
    def neutron(self):
        try:
            return neutron_client.Client(session=self.session.keystone_session)
        except neutron_exceptions.NeutronClientException as e:
            logger.exception('Failed to create neutron client: %s', e)
            six.reraise(OpenStackBackendError, e)

    @property
    def cinder(self):
        try:
            return cinder_client.Client(session=self.session.keystone_session)
        except cinder_exceptions.ClientException as e:
            logger.exception('Failed to create cinder client: %s', e)
            six.reraise(OpenStackBackendError, e)

    @property
    def glance(self):
        try:
            return glance_client.Client(session=self.session.keystone_session)
        except glance_exceptions.ClientException as e:
            logger.exception('Failed to create glance client: %s', e)
            six.reraise(OpenStackBackendError, e)

    @property
    def ceilometer(self):
        try:
            return ceilometer_client.Client('2', session=self.session.keystone_session)
        except ceilometer_exceptions.BaseException as e:
            logger.exception('Failed to create ceilometer client: %s', e)
            six.reraise(OpenStackBackendError, e)


class BaseOpenStackBackend(ServiceBackend):

    def __init__(self, settings, tenant_id=None):
        self.settings = settings
        self.tenant_id = tenant_id

    def _get_cached_session_key(self, admin):
        key = 'OPENSTACK_ADMIN_SESSION' if admin else 'OPENSTACK_SESSION_%s' % self.tenant_id
        settings_key = str(self.settings.backend_url) + str(self.settings.password) + str(self.settings.username)
        hashed_settings_key = hashlib.sha256(settings_key).hexdigest()
        return '%s_%s_%s' % (self.settings.uuid.hex, hashed_settings_key, key)

    def get_client(self, name=None, admin=False):
        domain_name = self.settings.domain or 'Default'
        credentials = {
            'auth_url': self.settings.backend_url,
            'username': self.settings.username,
            'password': self.settings.password,
            'user_domain_name': domain_name,
        }
        if self.tenant_id:
            credentials['project_id'] = self.tenant_id
        else:
            credentials['project_domain_name'] = domain_name
            credentials['project_name'] = self.settings.get_option('tenant_name')

        # Skip cache if service settings do no exist
        if not self.settings.uuid:
            return OpenStackClient(**credentials)

        client = None
        attr_name = 'admin_session' if admin else 'session'
        key = self._get_cached_session_key(admin)
        if hasattr(self, attr_name):  # try to get client from object
            client = getattr(self, attr_name)
        elif key in cache:  # try to get session from cache
            session = cache.get(key)
            try:
                client = OpenStackClient(session=session)
            except (OpenStackSessionExpired, OpenStackAuthorizationFailed):
                pass

        if client is None:  # create new token if session is not cached or expired
            client = OpenStackClient(**credentials)
            setattr(self, attr_name, client)  # Cache client in the object
            cache.set(key, dict(client.session), 24 * 60 * 60)  # Add session to cache

        if name:
            return getattr(client, name)
        else:
            return client

    def __getattr__(self, name):
        clients = 'keystone', 'nova', 'neutron', 'cinder', 'glance', 'ceilometer'
        for client in clients:
            if name == '{}_client'.format(client):
                return self.get_client(client, admin=False)

            if name == '{}_admin_client'.format(client):
                return self.get_client(client, admin=True)

        raise AttributeError(
            "'%s' object has no attribute '%s'" % (self.__class__.__name__, name))

    def ping(self, raise_exception=False):
        try:
            self.keystone_client
        except keystone_exceptions.ClientException as e:
            if raise_exception:
                six.reraise(OpenStackBackendError, e)
            return False
        else:
            return True

    def ping_resource(self, instance):
        try:
            self.nova_client.servers.get(instance.backend_id)
        except (ConnectionError, nova_exceptions.ClientException):
            return False
        else:
            return True

    def _pull_tenant_quotas(self, backend_id, scope):
        for quota_name, limit in self.get_tenant_quotas_limits(backend_id).items():
            scope.set_quota_limit(quota_name, limit)
        for quota_name, usage in self.get_tenant_quotas_usage(backend_id).items():
            scope.set_quota_usage(quota_name, usage, fail_silently=True)

    def get_tenant_quotas_limits(self, tenant_backend_id):
        nova = self.nova_client
        neutron = self.neutron_client
        cinder = self.cinder_client

        try:
            nova_quotas = nova.quotas.get(tenant_id=tenant_backend_id)
            cinder_quotas = cinder.quotas.get(tenant_id=tenant_backend_id)
            neutron_quotas = neutron.show_quota(tenant_id=tenant_backend_id)['quota']
        except (nova_exceptions.ClientException,
                cinder_exceptions.ClientException,
                neutron_exceptions.NeutronClientException) as e:
            six.reraise(OpenStackBackendError, e)

        return {
            Tenant.Quotas.ram: nova_quotas.ram,
            Tenant.Quotas.vcpu: nova_quotas.cores,
            Tenant.Quotas.storage: self.gb2mb(cinder_quotas.gigabytes),
            Tenant.Quotas.snapshots: cinder_quotas.snapshots,
            Tenant.Quotas.volumes: cinder_quotas.volumes,
            Tenant.Quotas.instances: nova_quotas.instances,
            Tenant.Quotas.security_group_count: neutron_quotas['security_group'],
            Tenant.Quotas.security_group_rule_count: neutron_quotas['security_group_rule'],
            Tenant.Quotas.floating_ip_count: neutron_quotas['floatingip'],
            Tenant.Quotas.network_count: neutron_quotas['network'],
            Tenant.Quotas.subnet_count: neutron_quotas['subnet'],
        }

    def get_tenant_quotas_usage(self, tenant_backend_id):
        nova = self.nova_client
        neutron = self.neutron_client
        cinder = self.cinder_client
        try:
            volumes = cinder.volumes.list()
            snapshots = cinder.volume_snapshots.list()
            instances = nova.servers.list()
            security_groups = neutron.list_security_groups(tenant_id=tenant_backend_id)['security_groups']
            floating_ips = neutron.list_floatingips(tenant_id=tenant_backend_id)['floatingips']
            networks = neutron.list_networks(tenant_id=tenant_backend_id)['networks']
            subnets = neutron.list_subnets(tenant_id=tenant_backend_id)['subnets']

            flavors = {flavor.id: flavor for flavor in nova.flavors.list()}

            ram, vcpu = 0, 0
            for flavor_id in (instance.flavor['id'] for instance in instances):
                try:
                    flavor = flavors.get(flavor_id, nova.flavors.get(flavor_id))
                except nova_exceptions.NotFound:
                    logger.warning('Cannot find flavor with id %s', flavor_id)
                    continue

                ram += getattr(flavor, 'ram', 0)
                vcpu += getattr(flavor, 'vcpus', 0)

        except (nova_exceptions.ClientException,
                cinder_exceptions.ClientException,
                neutron_exceptions.NeutronClientException) as e:
            six.reraise(OpenStackBackendError, e)

        volumes_size = sum(self.gb2mb(v.size) for v in volumes)
        snapshots_size = sum(self.gb2mb(v.size) for v in snapshots)
        storage = volumes_size + snapshots_size

        return {
            Tenant.Quotas.ram: ram,
            Tenant.Quotas.vcpu: vcpu,
            Tenant.Quotas.storage: storage,
            Tenant.Quotas.volumes: len(volumes),
            Tenant.Quotas.volumes_size: volumes_size,
            Tenant.Quotas.snapshots: len(snapshots),
            Tenant.Quotas.snapshots_size: snapshots_size,
            Tenant.Quotas.instances: len(instances),
            Tenant.Quotas.security_group_count: len(security_groups),
            Tenant.Quotas.security_group_rule_count: len(sum([sg['security_group_rules']
                                                              for sg in security_groups], [])),
            Tenant.Quotas.floating_ip_count: len(floating_ips),
            Tenant.Quotas.network_count: len(networks),
            Tenant.Quotas.subnet_count: len(subnets),
        }

    def _normalize_security_group_rule(self, rule):
        if rule['protocol'] is None:
            rule['protocol'] = ''

        if rule['remote_ip_prefix'] is None:
            rule['remote_ip_prefix'] = '0.0.0.0/0'

        if rule['port_range_min'] is None:
            rule['port_range_min'] = -1

        if rule['port_range_max'] is None:
            rule['port_range_max'] = -1

        return rule

    def _extract_security_group_rules(self, security_group, backend_security_group):
        backend_rules = backend_security_group['security_group_rules']
        cur_rules = {rule.backend_id: rule for rule in security_group.rules.all()}
        for backend_rule in backend_rules:
            # Currently we support only rules for incoming traffic
            if backend_rule['direction'] != 'ingress':
                continue
            cur_rules.pop(backend_rule['id'], None)
            backend_rule = self._normalize_security_group_rule(backend_rule)
            security_group.rules.update_or_create(
                backend_id=backend_rule['id'],
                defaults={
                    'from_port': backend_rule['port_range_min'],
                    'to_port': backend_rule['port_range_max'],
                    'protocol': backend_rule['protocol'],
                    'cidr': backend_rule['remote_ip_prefix'],
                })
        security_group.rules.filter(backend_id__in=cur_rules.keys()).delete()

    def _get_current_properties(self, model):
        return {p.backend_id: p for p in model.objects.filter(settings=self.settings)}

    def _pull_images(self, model_class, filter_function=None):
        glance = self.glance_client
        try:
            images = glance.images.list()
        except glance_exceptions.ClientException as e:
            six.reraise(OpenStackBackendError, e)

        images = [image for image in images if not image['status'] == 'deleted']
        if filter_function:
            images = filter(filter_function, images)

        with transaction.atomic():
            cur_images = self._get_current_properties(model_class)
            for backend_image in images:
                cur_images.pop(backend_image['id'], None)
                model_class.objects.update_or_create(
                    settings=self.settings,
                    backend_id=backend_image['id'],
                    defaults={
                        'name': backend_image['name'],
                        'min_ram': backend_image['min_ram'],
                        'min_disk': self.gb2mb(backend_image['min_disk']),
                    })
            model_class.objects.filter(backend_id__in=cur_images.keys(), settings=self.settings).delete()

    def _delete_backend_floating_ip(self, backend_id, tenant_backend_id):
        neutron = self.neutron_client
        try:
            logger.info("Deleting floating IP %s from tenant %s", backend_id, tenant_backend_id)
            neutron.delete_floatingip(backend_id)
        except neutron_exceptions.NotFound:
            logger.debug("Floating IP %s is already gone from tenant %s", backend_id, tenant_backend_id)
        except neutron_exceptions.NeutronClientException as e:
            six.reraise(OpenStackBackendError, e)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations


# Authentication method now uses API v3.
def update_service_settings_backend_url(apps, schema_editor):
    ServiceSettings = apps.get_model('structure', 'ServiceSettings')
    for service_settings in ServiceSettings.objects.filter(type__in=('OpenStack', 'OpenStackTenant')):
        if service_settings.backend_url.endswith('v2.0'):
            service_settings.backend_url = service_settings.backend_url[:-len('v2.0')] + 'v3'
            service_settings.save()


class Migration(migrations.Migration):

    dependencies = [
        ('structure', '0041_servicesettings_domain'),
        ('openstack', '0031_tenant_backup_storage'),
    ]

    operations = [
        migrations.RunPython(update_service_settings_backend_url),
    ]



import re

from cryptography import x509
from cryptography.hazmat.backends import default_backend
from django.db import transaction
from django.utils import six, timezone
from django.utils.translation import ugettext_lazy as _
from rest_framework import serializers
from rest_framework.reverse import reverse

from waldur_core.core import serializers as core_serializers
from waldur_core.structure import serializers as structure_serializers

from . import models
from .backend import AzureBackendError, SizeQueryset


class ServiceSerializer(core_serializers.ExtraFieldOptionsMixin,
                        structure_serializers.BaseServiceSerializer):

    SERVICE_ACCOUNT_FIELDS = {
        'username': _('In the format of GUID'),
        'certificate': _('X509 certificate in .PEM format'),
    }
    SERVICE_ACCOUNT_EXTRA_FIELDS = {
        'location': '',
        'cloud_service_name': '',
        'images_regex': ''
    }

    location = serializers.ChoiceField(
        choices=models.AzureService.Locations,
        write_only=True,
        required=False,
        allow_blank=True)

    class Meta(structure_serializers.BaseServiceSerializer.Meta):
        model = models.AzureService
        view_name = 'azure-detail'
        extra_field_options = {
            'username': {
                'label': 'Subscription ID',
                'required': True
            },
            'certificate': {
                'label': 'Private certificate file',
                'required': True,
                'write_only': True
            },
            'location': {
                'help_text': _('Azure region where to provision resources (default: "Central US")')
            },
            'cloud_service_name': {
                'help_text': _('Cloud service group to assign all connected SPLs to'),
                'required': True,
            },
            'images_regex': {
                'help_text': _('Regular expression to limit images list')
            }
        }

    def validate_certificate(self, value):
        if value:
            try:
                x509.load_pem_x509_certificate(value.read(), default_backend())
            except ValueError:
                raise serializers.ValidationError(_('Valid X509 certificate in .PEM format is expected'))

        return value


class ImageSerializer(structure_serializers.BasePropertySerializer):

    class Meta(object):
        model = models.Image
        view_name = 'azure-image-detail'
        fields = ('url', 'uuid', 'name')
        extra_kwargs = {
            'url': {'lookup_field': 'uuid'},
        }


class SizeSerializer(six.with_metaclass(structure_serializers.PropertySerializerMetaclass,
                                        serializers.Serializer)):

    uuid = serializers.ReadOnlyField()
    url = serializers.SerializerMethodField()
    name = serializers.ReadOnlyField()
    cores = serializers.ReadOnlyField()
    ram = serializers.ReadOnlyField()
    disk = serializers.ReadOnlyField()

    class Meta(object):
        model = models.Size

    def get_url(self, size):
        return reverse('azure-size-detail', kwargs={'uuid': size.uuid}, request=self.context.get('request'))


class ServiceProjectLinkSerializer(structure_serializers.BaseServiceProjectLinkSerializer):

    class Meta(structure_serializers.BaseServiceProjectLinkSerializer.Meta):
        model = models.AzureServiceProjectLink
        view_name = 'azure-spl-detail'
        extra_kwargs = {
            'service': {'lookup_field': 'uuid', 'view_name': 'azure-detail'},
        }


class InstanceEndpointsSerializer(serializers.ModelSerializer):
    local_port = serializers.ReadOnlyField()
    public_port = serializers.ReadOnlyField()
    protocol = serializers.ReadOnlyField()
    name = serializers.ReadOnlyField()

    class Meta(object):
        model = models.InstanceEndpoint
        fields = ('local_port', 'public_port', 'protocol', 'name')


class VirtualMachineSerializer(structure_serializers.BaseResourceSerializer):
    endpoints = InstanceEndpointsSerializer(many=True, read_only=True)

    service = serializers.HyperlinkedRelatedField(
        source='service_project_link.service',
        view_name='azure-detail',
        read_only=True,
        lookup_field='uuid')

    service_project_link = serializers.HyperlinkedRelatedField(
        view_name='azure-spl-detail',
        queryset=models.AzureServiceProjectLink.objects.all(),
        allow_null=True,
        required=False,
    )

    image = serializers.HyperlinkedRelatedField(
        view_name='azure-image-detail',
        lookup_field='uuid',
        queryset=models.Image.objects.all(),
        write_only=True)

    size = serializers.HyperlinkedRelatedField(
        view_name='azure-size-detail',
        lookup_field='uuid',
        queryset=SizeQueryset(),
        write_only=True)

    external_ips = serializers.ListField(
        child=serializers.IPAddressField(),
        read_only=True,
    )

    user_username = serializers.CharField(required=True)
    user_password = serializers.CharField(required=True, style={'input_type': 'password'})

    rdp = serializers.HyperlinkedIdentityField(view_name='azure-virtualmachine-rdp', lookup_field='uuid')

    class Meta(structure_serializers.BaseResourceSerializer.Meta):
        model = models.VirtualMachine
        view_name = 'azure-virtualmachine-detail'
        fields = structure_serializers.BaseResourceSerializer.Meta.fields + (
            'image', 'size', 'user_username', 'user_password', 'user_data', 'rdp', 'external_ips', 'internal_ips',
            'runtime_state', 'start_time', 'cores', 'ram', 'disk', 'image_name', 'endpoints',
        )
        protected_fields = structure_serializers.BaseResourceSerializer.Meta.protected_fields + (
            'image', 'size', 'user_username', 'user_password', 'user_data'
        )
        read_only_fields = structure_serializers.BaseResourceSerializer.Meta.read_only_fields + (
            'external_ips', 'internal_ips', 'runtime_state', 'start_time', 'cores', 'ram', 'disk',
            'image_name'
        )

    def validate(self, attrs):
        attrs = super(VirtualMachineSerializer, self).validate(attrs)

        if not re.match(r'[a-zA-Z][a-zA-Z0-9-]{0,13}[a-zA-Z0-9]$', attrs['name']):
            raise serializers.ValidationError(
                {'name': _("The name can contain only letters, numbers, and hyphens. "
                           "The name must be shorter than 15 characters and start with "
                           "a letter and must end with a letter or a number.")})

        # passwords must contain characters from at least three of the following four categories:
        groups = (r'[a-z]', r'[A-Z]', r'[0-9]', r'[^a-zA-Z\d\s:]')
        password = attrs['user_password']
        if not 6 <= len(password) <= 72 or sum(bool(re.search(g, password)) for g in groups) < 3:
            raise serializers.ValidationError({
                'user_password': _("The supplied password must be 6-72 characters long "
                                   "and contain 3 of the following: a lowercase character, "
                                   "an uppercase character, a number, a special character.")})

        if re.match(r'Administrator|Admin', attrs['user_username'], re.I):
            raise serializers.ValidationError({'user_username': _('Invalid administrator username.')})

        return attrs

    @transaction.atomic
    def create(self, validated_data):
        image = validated_data['image']
        validated_data['image_name'] = image.name
        size = validated_data['size']
        validated_data['cores'] = size.cores
        validated_data['ram'] = size.ram
        validated_data['disk'] = size.disk
        return super(VirtualMachineSerializer, self).create(validated_data)


class VirtualMachineImportSerializer(structure_serializers.BaseResourceImportSerializer):

    class Meta(structure_serializers.BaseResourceImportSerializer.Meta):
        model = models.VirtualMachine
        view_name = 'azure-virtualmachine-detail'
        fields = structure_serializers.BaseResourceImportSerializer.Meta.fields + (
            'cores', 'ram', 'disk',
            'external_ips', 'internal_ips',
        )

    def create(self, validated_data):
        spl = validated_data['service_project_link']
        backend = spl.get_backend()

        try:
            vm = backend.get_vm(validated_data['backend_id'])
        except AzureBackendError:
            raise serializers.ValidationError(
                {'backend_id': _("Can't find Virtual Machine with ID %s") % validated_data['backend_id']})

        validated_data['name'] = vm.name
        validated_data['created'] = timezone.now()
        validated_data['ram'] = vm.size.ram
        validated_data['disk'] = vm.size.disk
        validated_data['cores'] = 'Shared' and 1 or vm.size.extra['cores']
        validated_data['external_ips'] = vm.public_ips[0]
        validated_data['internal_ips'] = vm.private_ips[0]
        validated_data['state'] = models.VirtualMachine.States.ONLINE \
            if vm.state == backend.State.RUNNING else models.VirtualMachine.States.OFFLINE

        return super(VirtualMachineImportSerializer, self).create(validated_data)



from __future__ import unicode_literals

from celery import shared_task

from . import utils, models


@shared_task(name='marketplace.create_screenshot_thumbnail')
def create_screenshot_thumbnail(uuid):
    screenshot = models.Screenshots.objects.get(uuid=uuid)
    utils.create_screenshot_thumbnail(screenshot)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('cost_tracking', '0013_remove_item_type_choices'),
    ]

    operations = [
        migrations.AlterField(
            model_name='defaultpricelistitem',
            name='value',
            field=models.DecimalField(default=0, verbose_name=b'Hourly rate', max_digits=11, decimal_places=5),
            preserve_default=True,
        ),
        migrations.AlterField(
            model_name='pricelistitem',
            name='value',
            field=models.DecimalField(default=0, verbose_name=b'Hourly rate', max_digits=11, decimal_places=5),
            preserve_default=True,
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('core', '0003_user_registration_method'),
    ]

    operations = [
        migrations.AddField(
            model_name='user',
            name='agreement_date',
            field=models.DateTimeField(help_text='Indicates when the user has agreed with the policy.', null=True, verbose_name='agreement date', blank=True),
        ),
    ]



import errno
import logging
from shutil import rmtree

from celery import shared_task

logger = logging.getLogger(__name__)


@shared_task(name='waldur_ansible.tasks.delete_playbook_workspace')
def delete_playbook_workspace(workspace_path):
    logger.debug('Deleting playbook workspace %s.', workspace_path)
    try:
        rmtree(workspace_path)
    except OSError as e:
        if e.errno == errno.ENOENT:
            logger.info('Playbook workspace %s does not exist.', workspace_path)
        else:
            logger.warning('Failed to delete playbook workspace %s.', workspace_path)
            raise
    else:
        logger.info('Playbook workspace %s has been deleted.', workspace_path)



from __future__ import unicode_literals

import json
import requests
import uuid

from django.conf import settings
from django.contrib.auth import get_user_model
from django.db import transaction, IntegrityError
from rest_framework import views, status, response, generics
from rest_framework.authtoken.models import Token
from rest_framework.exceptions import ValidationError, APIException

from waldur_core.core.views import RefreshTokenMixin, validate_authentication_method

from . import tasks
from .log import event_logger, provider_event_type_mapping
from .models import AuthProfile
from .serializers import RegistrationSerializer, ActivationSerializer, AuthSerializer


auth_social_settings = getattr(settings, 'WALDUR_AUTH_SOCIAL', {})
GOOGLE_SECRET = auth_social_settings.get('GOOGLE_SECRET')
FACEBOOK_SECRET = auth_social_settings.get('FACEBOOK_SECRET')
SMARTIDEE_SECRET = auth_social_settings.get('SMARTIDEE_SECRET')

validate_social_signup = validate_authentication_method('SOCIAL_SIGNUP')
validate_local_signup = validate_authentication_method('LOCAL_SIGNUP')

User = get_user_model()


class AuthException(APIException):
    status_code = status.HTTP_401_UNAUTHORIZED


class FacebookException(AuthException):
    def __init__(self, facebook_error):
        self.message_text = facebook_error.get('message', 'Undefined')
        self.message_type = facebook_error.get('type', 'Undefined')
        self.message_code = facebook_error.get('code', 'Undefined')
        self.message = 'Facebook error {} (code:{}): {}'.format(self.message_type, self.message_code, self.message_text)
        super(FacebookException, self).__init__(detail=self.message)

    def __str__(self):
        return self.message


class GoogleException(AuthException):
    def __init__(self, google_error):
        if isinstance(google_error, basestring):
            self.message = google_error
        else:
            self.message_text = google_error.get('message', 'Undefined')
            self.message_code = google_error.get('code', 'Undefined')
            self.message = 'Google error (code:{}): {}'.format(self.message_code, self.message_text)

        super(GoogleException, self).__init__(detail=self.message)

    def __str__(self):
        return self.message


class SmartIDeeException(AuthException):
    def __init__(self, error_message, error_description):
        self.message = 'SmartIDee error: %s' % error_message
        if error_description:
            self.message = '%s (%s)' % (self.message, error_description)
        super(SmartIDeeException, self).__init__(detail=self.message)

    def __str__(self):
        return self.message


def generate_username(name):
    return uuid.uuid4().hex[:30]


class BaseAuthView(RefreshTokenMixin, views.APIView):
    permission_classes = []
    authentication_classes = []
    provider = None

    @validate_social_signup
    def post(self, request, format=None):
        if not self.request.user.is_anonymous:
            raise ValidationError('This view is for anonymous users only.')

        serializer = AuthSerializer(data={
            'client_id': request.data.get('clientId'),
            'redirect_uri': request.data.get('redirectUri'),
            'code': request.data.get('code')
        })
        serializer.is_valid(raise_exception=True)

        backend_user = self.get_backend_user(serializer.validated_data)
        user, created = self.create_or_update_user(backend_user)

        token = self.refresh_token(user)

        event_logger.auth_social.info(
            'User {user_username} with full name {user_full_name} authenticated successfully with {provider}.',
            event_type=provider_event_type_mapping[self.provider],
            event_context={
                'provider': self.provider,
                'user': user,
            }
        )
        return response.Response({'token': token.key},
                                 status=created and status.HTTP_201_CREATED or status.HTTP_200_OK)

    def get_backend_user(self, validated_data):
        """
        It should return dictionary with fields 'name' and 'id'
        """
        raise NotImplementedError

    def create_or_update_user(self, backend_user):
        user_id, user_name = backend_user['id'], backend_user['name']
        try:
            with transaction.atomic():
                user = User.objects.create_user(
                    username=generate_username(user_name),
                    full_name=user_name,
                    registration_method=self.provider
                )
                user.set_unusable_password()
                user.save()
                setattr(user.auth_profile, self.provider, user_id)
                user.auth_profile.save()
                return user, True
        except IntegrityError:
            profile = AuthProfile.objects.get(**{self.provider: user_id})
            if profile.user.full_name != user_name:
                profile.user.full_name = user_name
                profile.user.save()
            return profile.user, False


class GoogleView(BaseAuthView):

    provider = 'google'

    def get_backend_user(self, validated_data):
        access_token_url = 'https://www.googleapis.com/oauth2/v3/token'
        people_api_url = 'https://www.googleapis.com/plus/v1/people/me/openIdConnect'

        payload = dict(client_id=validated_data['client_id'],
                       redirect_uri=validated_data['redirect_uri'],
                       client_secret=GOOGLE_SECRET,
                       code=validated_data['code'],
                       grant_type='authorization_code')

        # Step 1. Exchange authorization code for access token.
        r = requests.post(access_token_url, data=payload)

        token = json.loads(r.text)
        self._assert_response_valid(token)
        headers = {'Authorization': 'Bearer {0}'.format(token['access_token'])}

        # Step 2. Retrieve information about the current user.
        r = requests.get(people_api_url, headers=headers)
        response_data = json.loads(r.text)

        # Step 3. Check is response valid.
        self._assert_response_valid(response_data)

        return {
            'id': response_data['sub'],
            'name': response_data['name']
        }

    def _assert_response_valid(self, response_data):
        if 'error' in response_data:
            raise GoogleException(response_data['error'])


class FacebookView(BaseAuthView):

    provider = 'facebook'

    def get_backend_user(self, validated_data):
        access_token_url = 'https://graph.facebook.com/oauth/access_token'
        graph_api_url = 'https://graph.facebook.com/me'

        params = {
            'client_id': validated_data['client_id'],
            'redirect_uri': validated_data['redirect_uri'],
            'client_secret': FACEBOOK_SECRET,
            'code': validated_data['code']
        }

        # Step 1. Exchange authorization code for access token.
        r = requests.get(access_token_url, params=params)
        self.check_response(r)
        params = {
            'access_token': r.json()['access_token']
        }

        # Step 2. Retrieve information about the current user.
        r = requests.get(graph_api_url, params=params)
        self.check_response(r)
        response_data = r.json()

        return {
            'id': response_data['id'],
            'name': response_data['name']
        }

    def check_response(self, r, valid_response=requests.codes.ok):
        if r.status_code != valid_response:
            try:
                data = r.json()
                error_message = data['error']
            except Exception:
                values = (r.reason, r.status_code)
                error_message = 'Message: %s, status code: %s' % values
            raise FacebookException(error_message)


class SmartIDeeView(BaseAuthView):
    provider = 'smartid.ee'

    def get_backend_user(self, validated_data):
        access_token_url = 'https://id.smartid.ee/oauth/access_token'
        user_data_url = 'https://id.smartid.ee/api/v2/user_data'

        data = {
            'client_id': validated_data['client_id'],
            'client_secret': SMARTIDEE_SECRET,
            'redirect_uri': validated_data['redirect_uri'],
            'code': validated_data['code'],
            'grant_type': 'authorization_code',
        }

        # Step 1. Exchange authorization code for access token.
        r = requests.post(access_token_url, data=data)
        self.check_response(r)
        access_token = r.json()['access_token']

        # Step 2. Retrieve information about the current user.
        r = requests.get(user_data_url, params={'access_token': access_token})
        self.check_response(r)
        return r.json()

    def check_response(self, r, valid_response=requests.codes.ok):
        if r.status_code != valid_response:
            try:
                data = r.json()
                error_message = data['error']
                error_description = data.get('error_description', '')
            except Exception:
                values = (r.reason, r.status_code)
                error_message = 'Message: %s, status code: %s' % values
                error_description = ''
            raise SmartIDeeException(error_message, error_description)

    def create_or_update_user(self, backend_user):
        """ Authenticate user by civil number """
        full_name = ('%s %s' % (backend_user['firstname'], backend_user['lastname']))[:100]
        try:
            user = User.objects.get(civil_number=backend_user['idcode'])
        except User.DoesNotExist:
            created = True
            user = User.objects.create_user(
                username=generate_username(full_name),
                # Ilja: disabling email update from smartid.ee as it comes in as a fake object for the moment.
                # email=backend_user['email'],
                full_name=full_name,
                civil_number=backend_user['idcode'],
                registration_method=self.provider,
            )
            user.set_unusable_password()
            user.save()
        else:
            created = False
            if user.full_name != full_name:
                user.full_name = full_name
                user.save()
        return user, created


class RegistrationView(generics.CreateAPIView):
    permission_classes = ()
    authentication_classes = ()
    serializer_class = RegistrationSerializer

    @validate_local_signup
    def post(self, request, *args, **kwargs):
        return super(RegistrationView, self).post(request, *args, **kwargs)

    def perform_create(self, serializer):
        user = serializer.save()
        user.is_active = False
        user.save()
        tasks.send_activation_email.delay(user.uuid.hex)


class ActivationView(views.APIView):
    permission_classes = ()
    authentication_classes = ()

    @validate_local_signup
    def post(self, request):
        serializer = ActivationSerializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        serializer.user.is_active = True
        serializer.user.save()

        token = Token.objects.get(user=serializer.user)
        return response.Response({'token': token.key}, status=status.HTTP_201_CREATED)



from waldur_core.logging.loggers import EventLogger, event_logger

from .models import Issue, Comment


class IssueEventLogger(EventLogger):
    issue = Issue

    class Meta:
        event_types = ('issue_deletion_succeeded',
                       'issue_update_succeeded',
                       'issue_creation_succeeded')
        event_groups = {
            'jira': event_types
        }


class CommentEventLogger(EventLogger):
    comment = Comment

    class Meta:
        event_types = ('comment_deletion_succeeded',
                       'comment_update_succeeded',
                       'comment_creation_succeeded')
        event_groups = {
            'jira': event_types
        }


event_logger.register('jira_issue', IssueEventLogger)
event_logger.register('jira_comment', CommentEventLogger)



import logging
from datetime import timedelta, datetime

from celery.task import Task as CeleryTask
from django.conf import settings
from django.utils import timezone

from waldur_core.core import tasks as core_tasks
from waldur_core.structure import SupportedServices

from . import models, executors

logger = logging.getLogger(__name__)


# TODO: Move this mixin to Waldur Core
class ExtensionTaskMixin(CeleryTask):
    """
    This mixin allows to skip task scheduling if extension is disabled.
    Subclasses should implement "is_extension_disabled" method which returns boolean value.
    """
    def is_extension_disabled(self):
        return False

    def apply_async(self, args=None, kwargs=None, **options):
        if self.is_extension_disabled():
            message = 'Task %s is not scheduled, because its extension is disabled.' % self.name
            logger.info(message)
            return self.AsyncResult(options.get('task_id'))
        return super(ExtensionTaskMixin, self).apply_async(args=args, kwargs=kwargs, **options)


class PaypalTaskMixin(ExtensionTaskMixin):
    def is_extension_disabled(self):
        return not settings.WALDUR_PAYPAL['ENABLED']


class DebitCustomers(PaypalTaskMixin, core_tasks.BackgroundTask):
    """ Fetch a list of shared services (services based on shared settings).
        Calculate the amount of consumed resources "yesterday" (make sure this task executed only once a day)
        Reduce customer's balance accordingly
        Stop online resource if needed
    """
    name = 'waldur_paypal.DebitCustomers'

    def is_equal(self, other_task, *args, **kwargs):
        return self.name == other_task.get('name')

    def run(self):
        date = datetime.now() - timedelta(days=1)
        start_date = date.replace(hour=0, minute=0, second=0, microsecond=0)
        end_date = start_date + timedelta(days=1, microseconds=-1)

        # XXX: it's just a placeholder, it doesn't work properly now nor implemented anyhow
        #      perhaps it should merely use price estimates..

        models = SupportedServices.get_resource_models().values()

        for model in models:
            resources = model.objects.filter(
                service_project_link__service__settings__shared=True)

            for resource in resources:
                try:
                    data = resource.get_cost(start_date, end_date)
                except NotImplementedError:
                    continue
                else:
                    resource.customer.debit_account(data['total_amount'])


class PaymentsCleanUp(PaypalTaskMixin, core_tasks.BackgroundTask):
    name = 'waldur_paypal.PaymentsCleanUp'

    def is_equal(self, other_task, *args, **kwargs):
        return self.name == other_task.get('name')

    def run(self):
        timespan = settings.WALDUR_PAYPAL.get('STALE_PAYMENTS_LIFETIME', timedelta(weeks=1))
        models.Payment.objects.filter(state=models.Payment.States.CREATED, created__lte=timezone.now() - timespan).delete()


class SendInvoices(PaypalTaskMixin, core_tasks.BackgroundTask):
    name = 'waldur_paypal.SendInvoices'

    def is_equal(self, other_task, *args, **kwargs):
        return self.name == other_task.get('name')

    def run(self):
        new_invoices = models.Invoice.objects.filter(backend_id='')

        for invoice in new_invoices.iterator():
            executors.InvoiceCreateExecutor.execute(invoice)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import django.core.validators
import django.utils.timezone
import model_utils.fields
from django.db import migrations, models

import waldur_core.core.fields
import waldur_core.core.validators
import waldur_core.cost_tracking.models


class Migration(migrations.Migration):
    replaces = [('cost_tracking', '0001_squashed_0012_fix_fields_length'),
                ('cost_tracking', '0001_initial'), ('cost_tracking', '0002_price_list'),
                ('cost_tracking', '0003_new_price_list_items'), ('cost_tracking', '0004_remove_connection_to_resource'),
                ('cost_tracking', '0005_expand_item_type_size'),
                ('cost_tracking', '0006_add_backend_cache_fields_to_pricelist'),
                ('cost_tracking', '0007_remove_obsolete_billing_fields'),
                ('cost_tracking', '0008_delete_resourceusage'), ('cost_tracking', '0009_defaultpricelistitem_name'),
                ('cost_tracking', '0010_applicationtype'), ('cost_tracking', '0011_applicationtype_slug'),
                ('cost_tracking', '0012_fix_fields_length'),
                ('cost_tracking', '0013_remove_item_type_choices'), ('cost_tracking', '0014_more_digits_for_price'),
                ('cost_tracking', '0015_defaultpricelistitem_metadata'),
                ('cost_tracking', '0016_leaf_estimates_squashed_0022_priceestimate_leafs'),
                ('cost_tracking', '0016_leaf_estimates'), ('cost_tracking', '0017_nullable_object_id'),
                ('cost_tracking', '0018_priceestimate_threshold'), ('cost_tracking', '0019_priceestimate_limit'),
                ('cost_tracking', '0020_reset_price_list_item'), ('cost_tracking', '0021_delete_applicationtype'),
                ('cost_tracking', '0022_priceestimate_leafs'),
                ('cost_tracking', '0023_consumptiondetails'), ('cost_tracking', '0024_refactor_price_estimate'),
                ('cost_tracking', '0025_increase_price_decimal_places'),
                ('cost_tracking', '0026_remove_limit_threshold')]

    initial = True

    dependencies = [
        ('structure', '0001_squashed_0054'),
        ('contenttypes', '0002_remove_content_type_name'),
        ('contenttypes', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='ConsumptionDetails',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('configuration', waldur_core.cost_tracking.models.ConsumableItemsField(default={}, help_text='Current resource configuration.')),
                ('last_update_time', models.DateTimeField(help_text='Last configuration change time.')),
                ('consumed_before_update', waldur_core.cost_tracking.models.ConsumableItemsField(default={}, help_text='How many consumables were used by resource before last update.')),
            ],
            options={
                'verbose_name': 'Consumption details',
                'verbose_name_plural': 'Consumption details',
            },
        ),
        migrations.CreateModel(
            name='DefaultPriceListItem',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=150, validators=[waldur_core.core.validators.validate_name], verbose_name='name')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('value', models.DecimalField(decimal_places=7, default=0, max_digits=13, verbose_name='Hourly rate')),
                ('units', models.CharField(blank=True, max_length=255)),
                ('item_type', models.CharField(help_text='Type of price list item. Examples: storage, flavor.', max_length=255)),
                ('key', models.CharField(help_text='Key that corresponds particular consumable. Example: name of flavor.', max_length=255)),
                ('metadata', waldur_core.core.fields.JSONField(blank=True, default={}, help_text='Details of the item, that corresponds price list item. Example: details of flavor.')),
                ('resource_content_type', models.ForeignKey(default=None, on_delete=django.db.models.deletion.CASCADE, to='contenttypes.ContentType')),
            ],
        ),
        migrations.CreateModel(
            name='PriceEstimate',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('object_id', models.PositiveIntegerField(null=True)),
                ('details', waldur_core.core.fields.JSONField(default={}, help_text='Saved scope details. Field is populated on scope deletion.')),
                ('total', models.FloatField(default=0, help_text='Predicted price for scope for current month.')),
                ('consumed', models.FloatField(default=0, help_text='Price for resource until now.')),
                ('month', models.PositiveSmallIntegerField(validators=[django.core.validators.MaxValueValidator(12), django.core.validators.MinValueValidator(1)])),
                ('year', models.PositiveSmallIntegerField()),
                ('content_type', models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE, related_name='+', to='contenttypes.ContentType')),
                ('parents', models.ManyToManyField(help_text='Price estimate parents', related_name='children', to='cost_tracking.PriceEstimate')),
            ],
            bases=(waldur_core.logging.loggers.LoggableMixin, models.Model, waldur_core.core.models.DescendantMixin),
        ),
        migrations.CreateModel(
            name='PriceListItem',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('value', models.DecimalField(decimal_places=7, default=0, max_digits=13, verbose_name='Hourly rate')),
                ('units', models.CharField(blank=True, max_length=255)),
                ('object_id', models.PositiveIntegerField()),
                ('content_type', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='contenttypes.ContentType')),
                ('default_price_list_item', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='cost_tracking.DefaultPriceListItem')),
            ],
        ),
        migrations.AddField(
            model_name='consumptiondetails',
            name='price_estimate',
            field=models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, related_name='consumption_details', to='cost_tracking.PriceEstimate'),
        ),
        migrations.AlterUniqueTogether(
            name='pricelistitem',
            unique_together=set([('content_type', 'object_id', 'default_price_list_item')]),
        ),
        migrations.AlterUniqueTogether(
            name='priceestimate',
            unique_together=set([('content_type', 'object_id', 'month', 'year')]),
        ),
        migrations.AlterUniqueTogether(
            name='defaultpricelistitem',
            unique_together=set([('key', 'item_type', 'resource_content_type')]),
        ),
    ]



from django.contrib.contenttypes.models import ContentType

from waldur_core.cost_tracking import models as cost_tracking_models


def get_service_price_list_items(service, resource_model):
    """ Return all price list item that belongs to given service """
    resource_content_type = ContentType.objects.get_for_model(resource_model)
    default_items = set(cost_tracking_models.DefaultPriceListItem.objects.filter(
        resource_content_type=resource_content_type))
    items = set(cost_tracking_models.PriceListItem.objects.filter(
        default_price_list_item__in=default_items, service=service).select_related('default_price_list_item'))
    rewrited_defaults = set([i.default_price_list_item for i in items])
    return items | (default_items - rewrited_defaults)



from waldur_ansible.python_management import models
from waldur_core.core import models as core_models


class NullErrorHandler(object):
    def handle_error(self, request, lines_post_processor):
        pass


class DeleteRequestErrorHandler(object):
    def handle_error(self, request, lines_post_processor):
        if not request.python_management.instance or self.exists_failed_initialization_request(request.python_management):
            request.python_management.delete()

    def exists_failed_initialization_request(self, python_management):
        return models.PythonManagementInitializeRequest.objects.get(python_management=python_management).state == core_models.StateMixin.States.ERRED



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models
import django.core.validators


class Migration(migrations.Migration):

    dependencies = [
        ('structure', '0037_remove_customer_billing_backend_id'),
        ('invoices', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='PaymentDetails',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('company', models.CharField(max_length=150, blank=True)),
                ('address', models.CharField(max_length=300, blank=True)),
                ('country', models.CharField(max_length=50, blank=True)),
                ('email', models.EmailField(max_length=75, blank=True)),
                ('postal', models.CharField(max_length=20, blank=True)),
                ('phone', models.CharField(max_length=20, blank=True)),
                ('bank', models.CharField(max_length=150, blank=True)),
                ('account', models.CharField(max_length=50, blank=True)),
                ('customer', models.OneToOneField(related_name='payment_details', to='structure.Customer')),
            ],
            options={'verbose_name': 'Payment details', 'verbose_name_plural': 'Payment details'},
        ),
        migrations.AddField(
            model_name='invoice',
            name='invoice_date',
            field=models.DateField(help_text='Date then invoice moved from state pending to created.', null=True, blank=True),
        ),
        migrations.AddField(
            model_name='invoice',
            name='tax_percent',
            field=models.DecimalField(default=0, max_digits=4, decimal_places=2, validators=[django.core.validators.MinValueValidator(0), django.core.validators.MaxValueValidator(100)]),
        ),
        migrations.AlterField(
            model_name='invoice',
            name='state',
            field=models.CharField(default='pending', max_length=30, choices=[('pending', 'Pending'), ('created', 'Created'), ('paid', 'Paid'), ('canceled', 'Canceled')]),
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion
import waldur_core.logging.loggers
import django.utils.timezone
import waldur_core.core.fields
import waldur_core.core.validators


class Migration(migrations.Migration):

    dependencies = [
        ('openstack_tenant', '0007_backup_backuprestoration'),
    ]

    operations = [
        migrations.CreateModel(
            name='BackupSchedule',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('description', models.CharField(max_length=500, verbose_name='description', blank=True)),
                ('name', models.CharField(max_length=150, verbose_name='name', validators=[waldur_core.core.validators.validate_name])),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('error_message', models.TextField(blank=True)),
                ('schedule', waldur_core.core.fields.CronScheduleField(max_length=15, validators=[waldur_core.core.validators.validate_cron_schedule, waldur_core.core.validators.MinCronValueValidator(1)])),
                ('next_trigger_at', models.DateTimeField(null=True)),
                ('timezone', models.CharField(default=django.utils.timezone.get_current_timezone_name, max_length=50)),
                ('is_active', models.BooleanField(default=False)),
                ('retention_time', models.PositiveIntegerField(help_text=b'Retention time in days, if 0 - backup will be kept forever')),
                ('maximal_number_of_backups', models.PositiveSmallIntegerField()),
                ('instance', models.ForeignKey(related_name='backup_schedules', to='openstack_tenant.Instance')),
            ],
            options={
                'abstract': False,
            },
            bases=(models.Model, waldur_core.logging.loggers.LoggableMixin),
        ),
        migrations.AddField(
            model_name='backup',
            name='backup_schedule',
            field=models.ForeignKey(related_name='backups', on_delete=django.db.models.deletion.SET_NULL, blank=True, to='openstack_tenant.BackupSchedule', null=True),
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models
import waldur_core.logging.loggers
import waldur_core.core.fields
import django.core.validators
import waldur_core.core.validators


class Migration(migrations.Migration):

    dependencies = [
        ('structure', '0037_remove_customer_billing_backend_id'),
        ('openstack_tenant', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='Flavor',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('name', models.CharField(max_length=150, verbose_name='name', validators=[waldur_core.core.validators.validate_name])),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('backend_id', models.CharField(max_length=255, db_index=True)),
                ('cores', models.PositiveSmallIntegerField(help_text=b'Number of cores in a VM')),
                ('ram', models.PositiveIntegerField(help_text=b'Memory size in MiB')),
                ('disk', models.PositiveIntegerField(help_text=b'Root disk size in MiB')),
                ('settings', models.ForeignKey(related_name='+', to='structure.ServiceSettings')),
            ],
            options={
                'abstract': False,
            },
            bases=(waldur_core.logging.loggers.LoggableMixin, models.Model),
        ),
        migrations.CreateModel(
            name='FloatingIP',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('name', models.CharField(max_length=150, verbose_name='name', validators=[waldur_core.core.validators.validate_name])),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('backend_id', models.CharField(max_length=255, db_index=True)),
                ('address', models.GenericIPAddressField(protocol=b'IPv4')),
                ('status', models.CharField(max_length=30)),
                ('backend_network_id', models.CharField(max_length=255, editable=False)),
                ('is_booked', models.BooleanField(default=False, help_text=b'Marks if floating IP has been booked for provisioning.')),
                ('settings', models.ForeignKey(related_name='+', to='structure.ServiceSettings')),
            ],
            options={
                'abstract': False,
                'verbose_name': 'Floating IP',
                'verbose_name_plural': 'Floating IPs',
            },
        ),
        migrations.CreateModel(
            name='Image',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('name', models.CharField(max_length=150, verbose_name='name', validators=[waldur_core.core.validators.validate_name])),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('backend_id', models.CharField(max_length=255, db_index=True)),
                ('min_disk', models.PositiveIntegerField(default=0, help_text=b'Minimum disk size in MiB')),
                ('min_ram', models.PositiveIntegerField(default=0, help_text=b'Minimum memory size in MiB')),
                ('settings', models.ForeignKey(related_name='+', to='structure.ServiceSettings')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='SecurityGroup',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('description', models.CharField(max_length=500, verbose_name='description', blank=True)),
                ('name', models.CharField(max_length=150, verbose_name='name', validators=[waldur_core.core.validators.validate_name])),
                ('uuid', waldur_core.core.fields.UUIDField()),
                ('backend_id', models.CharField(max_length=255, db_index=True)),
                ('settings', models.ForeignKey(related_name='+', to='structure.ServiceSettings')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='SecurityGroupRule',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('protocol', models.CharField(blank=True, max_length=4, choices=[(b'tcp', b'tcp'), (b'udp', b'udp'), (b'icmp', b'icmp')])),
                ('from_port', models.IntegerField(null=True, validators=[django.core.validators.MaxValueValidator(65535)])),
                ('to_port', models.IntegerField(null=True, validators=[django.core.validators.MaxValueValidator(65535)])),
                ('cidr', models.CharField(max_length=32, blank=True)),
                ('backend_id', models.CharField(max_length=128, blank=True)),
                ('security_group', models.ForeignKey(related_name='rules', to='openstack_tenant.SecurityGroup')),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.AlterUniqueTogether(
            name='image',
            unique_together=set([('settings', 'backend_id')]),
        ),
        migrations.AlterUniqueTogether(
            name='floatingip',
            unique_together=set([('settings', 'backend_id')]),
        ),
        migrations.AlterUniqueTogether(
            name='flavor',
            unique_together=set([('settings', 'backend_id')]),
        ),
    ]



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models
from django.conf import settings


class Migration(migrations.Migration):

    dependencies = [
        ('support', '0003_additional_issue_fields'),
    ]

    operations = [
        migrations.AlterField(
            model_name='issue',
            name='assignee',
            field=models.ForeignKey(related_name='issues', blank=True, to='support.SupportUser', help_text='Help desk user who will implement the issue', null=True),
        ),
        migrations.AlterField(
            model_name='issue',
            name='caller',
            field=models.ForeignKey(related_name='created_issues', to=settings.AUTH_USER_MODEL, help_text='Waldur user who has reported the issue.'),
        ),
        migrations.AlterField(
            model_name='issue',
            name='reporter',
            field=models.ForeignKey(related_name='reported_issues', blank=True, to='support.SupportUser', help_text='Help desk user who have created the issue that is reported by caller.', null=True),
        ),
    ]



from __future__ import unicode_literals

import copy
from xml.etree.ElementTree import iselement  # nosec

from defusedxml.ElementTree import parse as parse_xml, ParseError as XMLParseError
from libcloud.utils.py3 import httplib
from libcloud.common.azure import AzureServiceManagementConnection as _AzureServiceManagementConnection
from libcloud.common.azure import AzureResponse as _AzureResponse
from libcloud.compute.drivers.azure import AzureNodeDriver as _AzureNodeDriver
from libcloud.common.types import InvalidCredsError
from libcloud.common.types import LibcloudError, MalformedResponseError


def fixxpath(root, xpath):
    """ElementTree wants namespaces in its xpaths, so here we add them."""
    namespace, root_tag = root.tag[1:].split("}", 1)
    fixed_xpath = "/".join(["{%s}%s" % (namespace, e)
                            for e in xpath.split("/")])
    return fixed_xpath


def parse_error(body):
    code = body.findtext(fixxpath(body, 'Code'))
    message = body.findtext(fixxpath(body, 'Message'))
    message = message.split('\n')[0]
    error_msg = '%s: %s' % (code, message)
    return error_msg


"""
actual VM configs are not supported:
https://azure.microsoft.com/en-gb/pricing/details/virtual-machines/linux/

previous and currently supported virtual machine configs:
https://azure.microsoft.com/en-gb/pricing/details/virtual-machines/linux-previous/
"""
AZURE_COMPUTE_INSTANCE_TYPES = {
    'A0': {
        'id': 'ExtraSmall',
        'name': 'Extra Small Instance',
        'ram': 768,
        'disk': 20,
        'bandwidth': None,
        'price': '0.017',
        'max_data_disks': 1,
        'cores': 'Shared'
    },
    'A1': {
        'id': 'Small',
        'name': 'Small Instance',
        'ram': 1792,
        'disk': 70,
        'bandwidth': None,
        'price': '0.051',
        'max_data_disks': 2,
        'cores': 1
    },
    'A2': {
        'id': 'Medium',
        'name': 'Medium Instance',
        'ram': 3584,
        'disk': 135,
        'bandwidth': None,
        'price': '0.102',
        'max_data_disks': 4,
        'cores': 2
    },
    'A3': {
        'id': 'Large',
        'name': 'Large Instance',
        'ram': 7168,
        'disk': 285,
        'bandwidth': None,
        'price': '0.203',
        'max_data_disks': 8,
        'cores': 4
    },
    'A4': {
        'id': 'ExtraLarge',
        'name': 'Extra Large Instance',
        'ram': 14336,
        'disk': 605,
        'bandwidth': None,
        'price': '0.405',
        'max_data_disks': 16,
        'cores': 8
    },
    'A5': {
        'id': 'A5',
        'name': 'Memory Intensive Instance',
        'ram': 14336,
        'disk': 135,
        'bandwidth': None,
        'price': '0.228',
        'max_data_disks': 4,
        'cores': 2
    },
    'A6': {
        'id': 'A6',
        'name': 'A6 Instance',
        'ram': 28672,
        'disk': 285,
        'bandwidth': None,
        'price': '0.456',
        'max_data_disks': 8,
        'cores': 4
    },
    'A7': {
        'id': 'A7',
        'name': 'A7 Instance',
        'ram': 57344,
        'disk': 605,
        'bandwidth': None,
        'price': '0.911',
        'max_data_disks': 16,
        'cores': 8
    },
    'A8': {
        'id': 'A8',
        'name': 'A8 Instance',
        'ram': 57344,
        'disk': 382,
        'bandwidth': None,
        'price': '0.946',
        'max_data_disks': 16,
        'cores': 8
    },
    'A9': {
        'id': 'A9',
        'name': 'A9 Instance',
        'ram': 114688,
        'disk': 382,
        'bandwidth': None,
        'price': '1.892',
        'max_data_disks': 16,
        'cores': 16
    },
    'A10': {
        'id': 'A10',
        'name': 'A10 Instance',
        'ram': 57344,
        'disk': 382,
        'bandwidth': None,
        'price': '0.757',
        'max_data_disks': 16,
        'cores': 8
    },
    'A11': {
        'id': 'A11',
        'name': 'A11 Instance',
        'ram': 114688,
        'disk': 382,
        'bandwidth': None,
        'price': '1.513',
        'max_data_disks': 16,
        'cores': 16
    },
    'D1': {
        'id': 'Standard_D1',
        'name': 'D1 Faster Compute Instance',
        'ram': 3584,
        'disk': 50,
        'bandwidth': None,
        'price': '0.071',
        'max_data_disks': 2,
        'cores': 1
    },
    'D2': {
        'id': 'Standard_D2',
        'name': 'D2 Faster Compute Instance',
        'ram': 7168,
        'disk': 100,
        'bandwidth': None,
        'price': '0.142',
        'max_data_disks': 4,
        'cores': 2
    },
    'D3': {
        'id': 'Standard_D3',
        'name': 'D3 Faster Compute Instance',
        'ram': 14336,
        'disk': 200,
        'bandwidth': None,
        'price': '0.284',
        'max_data_disks': 8,
        'cores': 4
    },
    'D4': {
        'id': 'Standard_D4',
        'name': 'D4 Faster Compute Instance',
        'ram': 28672,
        'disk': 400,
        'bandwidth': None,
        'price': '0.567',
        'max_data_disks': 16,
        'cores': 8
    },
    'D11': {
        'id': 'Standard_D11',
        'name': 'D11 Faster Compute Instance',
        'ram': 14336,
        'disk': 100,
        'bandwidth': None,
        'price': '0.189',
        'max_data_disks': 4,
        'cores': 2
    },
    'D12': {
        'id': 'Standard_D12',
        'name': 'D12 Faster Compute Instance',
        'ram': 28672,
        'disk': 200,
        'bandwidth': None,
        'price': '0.379',
        'max_data_disks': 8,
        'cores': 4
    },
    'D13': {
        'id': 'Standard_D13',
        'name': 'D13 Faster Compute Instance',
        'ram': 57344,
        'disk': 400,
        'bandwidth': None,
        'price': '0.757',
        'max_data_disks': 16,
        'cores': 8
    },
    'D14': {
        'id': 'Standard_D14',
        'name': 'D14 Faster Compute Instance',
        'ram': 114688,
        'disk': 800,
        'bandwidth': None,
        'price': '1.492',
        'max_data_disks': 32,
        'cores': 16
    }
}


class AzureResponse(_AzureResponse):
    """
    Fix error parsing for Azure
    """

    def parse_error(self, msg=None):
        error_msg = 'Unknown error'

        try:
            # Azure does give some meaningful errors, but is inconsistent
            # Some APIs respond with an XML error. Others just dump HTML
            body = self.parse_body()

            if iselement(body):
                error_msg = parse_error(body)

        except MalformedResponseError:
            pass

        if msg:
            error_msg = '%s - %s' % (msg, error_msg)

        if self.status in [httplib.UNAUTHORIZED, httplib.FORBIDDEN]:
            raise InvalidCredsError(error_msg)

        raise LibcloudError(
            '%s Status code: %d.' % (error_msg, self.status),
            driver=self
        )


class AzureServiceManagementConnection(_AzureServiceManagementConnection):
    responseCls = AzureResponse


class AzureNodeDriver(_AzureNodeDriver):
    connectionCls = AzureServiceManagementConnection

    def raise_for_response(self, response, valid_response):
        if response.status != valid_response:
            error_msg = response.body
            try:
                body = parse_xml(error_msg)
                if iselement(body):
                    error_msg = parse_error(body)
            except XMLParseError:
                pass

            values = (response.error, error_msg, response.status)
            message = 'Message: %s, Body: %s, Status code: %s' % (values)
            raise LibcloudError(message, driver=self)

    def _parse_response_body_from_xml_text(self, response, return_type):
        """
        parse the xml and fill all the data into a class of return_type
        """
        # required to support libcloud v[2.0, 2.1]
        if isinstance(response.body, unicode):
            response.body = response.body.encode('utf-8')

        return super(AzureNodeDriver, self)._parse_response_body_from_xml_text(response, return_type)

    def list_sizes(self):
        """
        Replaces AzureNodeDriver's list_sizes due to price change in Azure.
        """
        sizes = []

        for _, values in AZURE_COMPUTE_INSTANCE_TYPES.items():
            node_size = self._to_node_size(copy.deepcopy(values))
            sizes.append(node_size)

        return sizes



# -*- coding: utf-8 -*-
# Generated by Django 1.11.7 on 2018-06-29 12:52
from __future__ import unicode_literals

import django.core.validators
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('marketplace', '0008_native'),
    ]

    operations = [
        migrations.AlterField(
            model_name='offering',
            name='rating',
            field=models.IntegerField(help_text='Rating is value from 1 to 5.', null=True, validators=[django.core.validators.MaxValueValidator(5), django.core.validators.MinValueValidator(1)]),
        ),
    ]



import logging

from rest_framework import exceptions


logger = logging.getLogger(__name__)


class PluginManager(object):
    def __init__(self):
        self.backends = {}

    def register(self, model_class, processor):
        self.backends[model_class] = processor

    def get_processor(self, model_class):
        return self.backends.get(model_class)

    def process(self, order_item, request):
        processor = self.get_processor(order_item.offering.scope._meta.model)
        if not processor:
            raise exceptions.ValidationError('Skipping order item processing because processor is not found.')

        processor(order_item, request)


manager = PluginManager()



from __future__ import unicode_literals

from .log import event_logger


def log_openstack_package_creation(sender, instance, created=False, **kwargs):
    if not created:
        return

    event_logger.openstack_package.info(
        'OpenStack package {tenant_name} has been created.',
        event_type='openstack_package_created',
        event_context={
            'tenant': instance.tenant,
            'package_template_name': instance.template.name,
            'service_settings': instance.service_settings,
        })


def log_openstack_package_deletion(sender, instance, **kwargs):
    event_logger.openstack_package.info(
        'OpenStack package {tenant_name} has been deleted.',
        event_type='openstack_package_deleted',
        event_context={
            'tenant': instance.tenant,
            'package_template_name': instance.template.name,
            'service_settings': instance.service_settings,
        })



from datetime import timedelta

from django import forms
from django.conf import settings
from django.db import transaction
from django.utils import timezone
from rest_framework import serializers, exceptions

from waldur_core.core.fields import MappedChoiceField
from waldur_core.core.serializers import GenericRelatedField, HyperlinkedRelatedModelSerializer
from waldur_core.core.utils import datetime_to_timestamp, pwgen
from waldur_core.monitoring.utils import get_period
from waldur_core.structure import serializers as structure_serializers, models as structure_models

from . import models, apps


class ServiceSerializer(structure_serializers.BaseServiceSerializer):
    SERVICE_ACCOUNT_FIELDS = {
        'backend_url': 'Zabbix API URL (e.g. http://example.com/zabbix/api_jsonrpc.php)',
        'username': 'Zabbix user username (e.g. admin)',
        'password': 'Zabbix user password (e.g. zabbix)',
    }
    SERVICE_ACCOUNT_EXTRA_FIELDS = {
        'host_group_name': 'Zabbix host group name for registered hosts',
        'templates_names': 'List of Zabbix hosts templates',
        'database_parameters': 'Zabbix database parameters',
        'interface_parameters': 'Default parameters for hosts interface (will be used if interface is not specified)',
    }

    class Meta(structure_serializers.BaseServiceSerializer.Meta):
        model = models.ZabbixService
        view_name = 'zabbix-detail'
        required_fields = ('backend_url', 'username', 'password')


class ServiceProjectLinkSerializer(structure_serializers.BaseServiceProjectLinkSerializer):
    class Meta(structure_serializers.BaseServiceProjectLinkSerializer.Meta):
        model = models.ZabbixServiceProjectLink
        view_name = 'zabbix-spl-detail'
        extra_kwargs = {
            'service': {'lookup_field': 'uuid', 'view_name': 'zabbix-detail'},
        }


class TemplateSerializer(structure_serializers.BasePropertySerializer):
    items = serializers.SerializerMethodField()
    triggers = serializers.SerializerMethodField()

    class Meta(structure_serializers.BasePropertySerializer.Meta):
        model = models.Template
        view_name = 'zabbix-template-detail'
        fields = ('url', 'uuid', 'name', 'items', 'triggers', 'settings', 'children', 'parents')
        extra_kwargs = {
            'url': {'lookup_field': 'uuid'},
            'children': {'lookup_field': 'uuid', 'view_name': 'zabbix-template-detail'},
            'parents': {'lookup_field': 'uuid', 'view_name': 'zabbix-template-detail'},
            'settings': {'lookup_field': 'uuid'},
        }

    def get_items(self, template):
        items = template.items.all().values('name', 'key', 'units', 'value_type')
        for item in items:  # replace value types with human-readable names
            item['value_type'] = dict(models.Item.ValueTypes.CHOICES)[item['value_type']]
        return items

    def get_triggers(self, template):
        return template.triggers.all().values_list('name', flat=True)


class NestedTemplateSerializer(TemplateSerializer, HyperlinkedRelatedModelSerializer):
    class Meta(TemplateSerializer.Meta):
        pass


class HostSerializer(structure_serializers.BaseResourceSerializer):
    service = serializers.HyperlinkedRelatedField(
        source='service_project_link.service',
        view_name='zabbix-detail',
        read_only=True,
        lookup_field='uuid')

    service_project_link = serializers.HyperlinkedRelatedField(
        view_name='zabbix-spl-detail',
        queryset=models.ZabbixServiceProjectLink.objects.all(),
        allow_null=True,
        required=False,
    )

    # visible name could be populated from scope, so we need to mark it as not required
    visible_name = serializers.CharField(required=False, max_length=models.Host.VISIBLE_NAME_MAX_LENGTH)
    scope = GenericRelatedField(related_models=structure_models.ResourceMixin.get_all_models(), required=False)
    templates = NestedTemplateSerializer(
        queryset=models.Template.objects.all().prefetch_related('items', 'children'), many=True, required=False)
    status = MappedChoiceField(
        choices={v: v for _, v in models.Host.Statuses.CHOICES},
        choice_mappings={v: k for k, v in models.Host.Statuses.CHOICES},
        read_only=True,
    )
    interface_ip = serializers.IPAddressField(allow_blank=True, required=False, write_only=True,
                                              help_text='IP of host interface.')
    interface_parameters = serializers.JSONField(read_only=True)

    class Meta(structure_serializers.BaseResourceSerializer.Meta):
        model = models.Host
        view_name = 'zabbix-host-detail'
        fields = structure_serializers.BaseResourceSerializer.Meta.fields + (
            'visible_name', 'host_group_name', 'scope', 'templates', 'error', 'status',
            'interface_ip', 'interface_parameters',)
        read_only_fields = structure_serializers.BaseResourceSerializer.Meta.read_only_fields + (
            'error', 'interface_parameters')
        protected_fields = structure_serializers.BaseResourceSerializer.Meta.protected_fields + (
            'interface_ip', 'visible_name')

    def get_resource_fields(self):
        return super(HostSerializer, self).get_resource_fields() + ['scope']

    def validate(self, attrs):
        attrs = super(HostSerializer, self).validate(attrs)

        # model validation
        if self.instance is not None:
            for name, value in attrs.items():
                setattr(self.instance, name, value)
            self.instance.clean()
        else:
            service_settings = attrs.get('service_project_link').service.settings
            if service_settings.state == structure_models.ServiceSettings.States.ERRED:
                raise serializers.ValidationError('It is impossible to create host if service is in ERRED state.')

            if not attrs.get('visible_name'):
                if 'scope' not in attrs:
                    raise serializers.ValidationError('Visible name or scope should be defined.')

                # initiate name and visible name from scope if it is defined
                attrs['visible_name'] = models.Host.get_visible_name_from_scope(attrs['scope'])

            spl = attrs['service_project_link']
            if models.Host.objects.filter(
                    service_project_link__service__settings=spl.service.settings,
                    visible_name=attrs['visible_name']
            ).exists():
                raise serializers.ValidationError({'visible_name': 'Visible name should be unique.'})

            instance = models.Host(**{k: v for k, v in attrs.items() if k not in ('templates', 'interface_ip')})
            instance.clean()

        spl = attrs.get('service_project_link') or self.instance.service_project_link
        templates = attrs.get('templates', [])
        parents = {}  # dictionary <parent template: child template>
        for template in templates:
            if template.settings != spl.service.settings:
                raise serializers.ValidationError(
                    {'templates': 'Template "%s" and host belong to different service settings.' % template.name})
            for child in template.children.all():
                if child in templates:
                    message = 'Template "%s" is already registered as child of template "%s"' % (
                        child.name, template.name)
                    raise serializers.ValidationError({'templates': message})
            for parent in template.parents.all():
                if parent in parents:
                    message = 'Templates %s and %s belong to the same parent %s' % (template, parents[parent], parent)
                    raise serializers.ValidationError({'templates': message})
                else:
                    parents[parent] = template

        for template in templates:
            if template in parents:
                message = 'Template "%s" is already registered as a parent of template "%s"' % \
                          (template, parents[template])
                raise serializers.ValidationError({'templates': message})

        return attrs

    def create(self, validated_data):
        # define interface parameters based on settings and user input
        spl = validated_data['service_project_link']
        interface_parameters = spl.service.settings.get_option('interface_parameters')
        scope = validated_data.get('scope')
        interface_ip = validated_data.pop('interface_ip', None) or getattr(scope, 'internal_ips', None)
        if interface_ip:
            # Note, that we're not supporting multiple network interfaces for single host yet.
            # IPv6 is not supported yet too.
            if isinstance(interface_ip, list):
                interface_ip = interface_ip[0]
            interface_parameters['ip'] = interface_ip
        validated_data['interface_parameters'] = interface_parameters

        # populate templates
        templates = validated_data.pop('templates', None)
        with transaction.atomic():
            host = super(HostSerializer, self).create(validated_data)
            # get default templates from service settings if they are not defined
            if templates is None:
                templates = models.Template.objects.filter(
                    settings=host.service_project_link.service.settings,
                    name__in=host.service_project_link.service.settings.get_option('templates_names'),
                )
            for template in templates:
                host.templates.add(template)

        return host

    def update(self, host, validated_data):
        templates = validated_data.pop('templates', None)
        with transaction.atomic():
            host = super(HostSerializer, self).update(host, validated_data)
            if templates is not None:
                host.templates.clear()
                for template in templates:
                    host.templates.add(template)

        return host


class ITServiceSerializer(structure_serializers.BaseResourceSerializer):
    service = serializers.HyperlinkedRelatedField(
        source='service_project_link.service',
        view_name='zabbix-detail',
        read_only=True,
        lookup_field='uuid')

    service_project_link = serializers.HyperlinkedRelatedField(
        view_name='zabbix-spl-detail',
        queryset=models.ZabbixServiceProjectLink.objects.all(),
        allow_null=True,
        required=False,
    )

    host = serializers.HyperlinkedRelatedField(
        view_name='zabbix-host-detail',
        queryset=models.Host.objects.all(),
        lookup_field='uuid')

    trigger = serializers.HyperlinkedRelatedField(
        view_name='zabbix-trigger-detail',
        queryset=models.Trigger.objects.order_by('name').select_related('settings'),
        lookup_field='uuid')

    algorithm = MappedChoiceField(
        choices={v: v for _, v in models.ITService.Algorithm.CHOICES},
        choice_mappings={v: k for k, v in models.ITService.Algorithm.CHOICES},
    )
    trigger_name = serializers.ReadOnlyField(source='trigger.name')
    actual_sla = serializers.SerializerMethodField()

    class Meta(structure_serializers.BaseResourceSerializer.Meta):
        model = models.ITService
        view_name = 'zabbix-itservice-detail'
        fields = structure_serializers.BaseResourceSerializer.Meta.fields + (
            'host', 'algorithm', 'sort_order', 'agreed_sla', 'actual_sla', 'trigger', 'trigger_name', 'is_main')

    # XXX: Should we display sla here?
    def get_actual_sla(self, itservice):
        key = 'itservice_sla_map'
        if key not in self.context:
            qs = models.SlaHistory.objects.filter(period=get_period(self.context['request']))
            if isinstance(self.instance, list):
                qs = qs.filter(itservice__in=self.instance)
            else:
                qs = qs.filter(itservice=self.instance)
            self.context[key] = {q.itservice_id: q.value for q in qs}

        return self.context[key].get(itservice.id)

    def validate(self, attrs):
        attrs = super(ITServiceSerializer, self).validate(attrs)

        host = attrs.get('host')
        if host:
            trigger = attrs['trigger']

            if host and not host.templates.filter(id=trigger.template_id).exists():
                raise serializers.ValidationError("Host templates should contain trigger's template")

            if host.service_project_link != attrs['service_project_link']:
                raise serializers.ValidationError('Host and IT service should belong to the same SPL.')

        return attrs


class TriggerSerializer(structure_serializers.BasePropertySerializer):
    template = serializers.HyperlinkedRelatedField(
        view_name='zabbix-template-detail',
        read_only=True,
        lookup_field='uuid')
    priority = serializers.IntegerField()

    class Meta(structure_serializers.BasePropertySerializer.Meta):
        model = models.Trigger
        fields = ('url', 'uuid', 'name', 'priority', 'template')
        extra_kwargs = {
            'url': {'lookup_field': 'uuid', 'view_name': 'zabbix-trigger-detail'},
        }


class SlaHistoryEventSerializer(serializers.Serializer):
    timestamp = serializers.IntegerField()
    state = serializers.CharField()


class ItemsAggregatedValuesSerializer(serializers.Serializer):
    """ Validate input parameters for items_aggregated_values action. """
    start = serializers.IntegerField(default=lambda: datetime_to_timestamp(timezone.now() - timedelta(hours=1)))
    end = serializers.IntegerField(default=lambda: datetime_to_timestamp(timezone.now()))
    method = serializers.ChoiceField(default='MAX', choices=('MIN', 'MAX'))

    def validate(self, data):
        """
        Check that the start is before the end.
        """
        if 'start' in data and 'end' in data and data['start'] >= data['end']:
            raise serializers.ValidationError("End must occur after start")
        return data


class UserGroupSerializer(structure_serializers.BasePropertySerializer):
    class Meta(object):
        model = models.UserGroup
        fields = 'url', 'name', 'settings'
        read_only_fields = 'url', 'backend_id'
        extra_kwargs = {
            'url': {'lookup_field': 'uuid', 'view_name': 'zabbix-user-group-detail'},
            'settings': {'lookup_field': 'uuid'},
        }


class NestedUserGroupSerializer(UserGroupSerializer, HyperlinkedRelatedModelSerializer):
    class Meta(UserGroupSerializer.Meta):
        pass


class UserSerializer(structure_serializers.BasePropertySerializer):
    groups = NestedUserGroupSerializer(queryset=models.UserGroup.objects.all(), many=True)
    state = MappedChoiceField(
        choices={v: v for _, v in models.User.States.CHOICES},
        choice_mappings={v: k for k, v in models.User.States.CHOICES},
        read_only=True,
    )
    type = MappedChoiceField(
        choices={v: v for _, v in models.User.Types.CHOICES},
        choice_mappings={v: k for k, v in models.User.Types.CHOICES},
    )
    password = serializers.SerializerMethodField(
        help_text='Password is visible only after user creation or if user has type "default".')

    class Meta(object):
        model = models.User
        fields = ('url', 'alias', 'name', 'surname', 'type', 'groups', 'backend_id', 'settings', 'state', 'phone',
                  'password')
        read_only_fields = ('url', 'backend_id',)
        protected_fields = ('settings',)
        extra_kwargs = {
            'url': {'lookup_field': 'uuid', 'view_name': 'zabbix-user-detail'},
            'settings': {'lookup_field': 'uuid'},
        }

    def get_password(self, user):
        show_password = ((self.context['request'].method == 'POST' and user is None) or
                         user.type == models.User.Types.DEFAULT)
        return user.password if show_password else None

    def get_fields(self):
        fields = super(UserSerializer, self).get_fields()
        fields['settings'].queryset = structure_models.ServiceSettings.objects.filter(
            type=apps.ZabbixConfig.service_name)
        return fields

    def validate_type(self, value):
        user = self.context['request'].user
        if not user.is_staff and value != models.User.Types.DEFAULT:
            raise serializers.ValidationError('Cannot create not default user.')
        return value

    def validate(self, attrs):
        settings = attrs.get('settings') or self.instance.settings
        groups = attrs.get('groups', [])
        if any([group.settings != settings for group in groups]):
            raise serializers.ValidationError('User groups and user should belong to the same service settings')
        return attrs

    def create(self, attrs):
        groups = attrs.pop('groups', [])
        attrs['password'] = pwgen()
        user = super(UserSerializer, self).create(attrs)
        user.groups.add(*groups)
        return user

    def update(self, user, attrs):
        new_groups = set(attrs.pop('groups', []))
        old_groups = set(user.groups.all())
        user = super(UserSerializer, self).update(user, attrs)
        user.groups.remove(*(old_groups - new_groups))
        user.groups.add(*(new_groups - old_groups))
        return user


class TriggerRequestSerializer(serializers.Serializer):
    changed_before = serializers.DateTimeField(required=False)
    changed_after = serializers.DateTimeField(required=False)
    min_priority = serializers.ChoiceField(choices=models.Trigger.Priority.CHOICES, required=False)
    priority = serializers.MultipleChoiceField(choices=models.Trigger.Priority.CHOICES, required=False)
    acknowledge_status = serializers.ChoiceField(choices=models.Trigger.AcknowledgeStatus.CHOICES, required=False)
    host_name = serializers.CharField(required=False)
    host_id = serializers.CharField(required=False)
    # Value is not a good name for the filter, but let's keep consistency with Zabbix API.
    value = serializers.ChoiceField(choices=models.Trigger.Value.CHOICES, required=False)

    def validate(self, attrs):
        self._add_field_from_initial_data(attrs, 'include_events_count')
        self._add_field_from_initial_data(attrs, 'include_trigger_hosts')
        return attrs

    def _add_field_from_initial_data(self, attrs, name):
        param = self.initial_data.get(name)
        boolean_field = forms.NullBooleanField()
        try:
            param = boolean_field.to_python(param)
        except exceptions.ValidationError:
            param = None

        attrs[name] = param


class TriggerResponseSerializer(serializers.Serializer):
    changed = serializers.DateTimeField()
    hosts = serializers.ReadOnlyField()
    event_count = serializers.IntegerField()

    def get_fields(self):
        fields = super(TriggerResponseSerializer, self).get_fields()
        for field in settings.WALDUR_ZABBIX['TRIGGER_FIELDS']:
            fields[field[0]] = getattr(serializers, field[2])()
        return fields



from django import forms
from django.contrib import admin
from django.core import urlresolvers

from . import models


class CachedRepositoryPythonLibraryAdminForm(forms.ModelForm):
    class Meta:
        model = models.CachedRepositoryPythonLibrary
        fields = ('name',)


class CachedRepositoryPythonLibraryAdmin(admin.ModelAdmin):
    list_filter = ('name',)
    list_display = ('name',)


class PythonManagementAdminForm(forms.ModelForm):
    class Meta:
        model = models.PythonManagement
        fields = ('system_user', 'project', 'user')
        readonly_fields = ('uuid', 'created', 'virtual_envs_dir_path', 'instance_content_type', 'instance_object_id', 'python_version')


class PythonManagementAdmin(admin.ModelAdmin):
    list_filter = ('virtual_envs_dir_path', 'python_version', 'system_user', 'project', 'user')
    list_display = ('uuid', 'created', 'instance_content_type', 'instance_object_id', 'virtual_envs_dir_path', 'python_version', 'system_user', 'project', 'user')
    list_display_links = ('uuid',)


class RequestAdminForm(forms.ModelForm):
    class Meta:
        fields = ('state', 'output')
        readonly_fields = ('uuid', 'created', 'output')


class RequestAdmin(admin.ModelAdmin):
    list_display = ('uuid', 'created', 'state', 'python_management_link')
    list_display_links = ('uuid',)

    def python_management_link(self, obj):
        link = urlresolvers.reverse("admin:python_management_pythonmanagement_change", args=[obj.id])
        return '<a href="%s">%s</a>' % (link, obj.python_management.virtual_envs_dir_path)

    python_management_link.allow_tags = True
    python_management_link.short_description = 'Python Management'


class PythonManagementInitializeRequestAdminForm(RequestAdminForm):
    class Meta(RequestAdminForm.Meta):
        model = models.PythonManagementInitializeRequest


class PythonManagementInitializeRequestAdmin(RequestAdmin):
    pass


class PythonManagementSynchronizeRequestAdminForm(RequestAdminForm):
    class Meta(RequestAdminForm.Meta):
        model = models.PythonManagementSynchronizeRequest


class PythonManagementSynchronizeRequestAdmin(RequestAdmin):
    pass


class PythonManagementDeleteRequestAdminForm(RequestAdminForm):
    class Meta(RequestAdminForm.Meta):
        model = models.PythonManagementDeleteRequest


class PythonManagementDeleteRequestAdmin(RequestAdmin):
    pass


class PythonManagementDeleteVirtualEnvRequestAdminForm(RequestAdminForm):
    class Meta(RequestAdminForm.Meta):
        model = models.PythonManagementDeleteVirtualEnvRequest


class PythonManagementDeleteVirtualEnvRequestAdmin(RequestAdmin):
    pass


class PythonManagementFindVirtualEnvsRequestAdminForm(RequestAdminForm):
    class Meta(RequestAdminForm.Meta):
        model = models.PythonManagementFindVirtualEnvsRequest


class PythonManagementFindVirtualEnvsRequestAdmin(RequestAdmin):
    pass


class PythonManagementFindInstalledLibrariesRequestAdminForm(RequestAdminForm):
    class Meta(RequestAdminForm.Meta):
        model = models.PythonManagementFindInstalledLibrariesRequest


class PythonManagementFindInstalledLibrariesRequestAdmin(RequestAdmin):
    pass


admin.site.register(models.CachedRepositoryPythonLibrary, CachedRepositoryPythonLibraryAdmin)
admin.site.register(models.PythonManagement, PythonManagementAdmin)
admin.site.register(models.PythonManagementInitializeRequest, PythonManagementInitializeRequestAdmin)
admin.site.register(models.PythonManagementSynchronizeRequest, PythonManagementSynchronizeRequestAdmin)
admin.site.register(models.PythonManagementDeleteRequest, PythonManagementDeleteRequestAdmin)
admin.site.register(models.PythonManagementDeleteVirtualEnvRequest, PythonManagementDeleteVirtualEnvRequestAdmin)
admin.site.register(models.PythonManagementFindVirtualEnvsRequest, PythonManagementFindVirtualEnvsRequestAdmin)
admin.site.register(models.PythonManagementFindInstalledLibrariesRequest, PythonManagementFindInstalledLibrariesRequestAdmin)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('openstack', '0012_require_floating_ip_tenant'),
    ]

    operations = [
        migrations.AddField(
            model_name='backupschedule',
            name='backup_type',
            field=models.CharField(default='Regular', max_length=30, choices=[('Regular', 'Regular'), ('DR', 'DR')]),
        ),
        migrations.AddField(
            model_name='backupschedule',
            name='error_message',
            field=models.TextField(blank=True),
        ),
        migrations.AddField(
            model_name='backupschedule',
            name='runtime_state',
            field=models.CharField(max_length=150, verbose_name='runtime state', blank=True),
        ),
        migrations.AddField(
            model_name='drbackup',
            name='backup_schedule',
            field=models.ForeignKey(related_name='dr_backups', on_delete=django.db.models.deletion.SET_NULL, blank=True, to='openstack.BackupSchedule', null=True),
        ),
        migrations.AddField(
            model_name='drbackup',
            name='kept_until',
            field=models.DateTimeField(help_text='Guaranteed time of backup retention. If null - keep forever.', null=True, blank=True),
        ),
        migrations.AlterField(
            model_name='backupschedule',
            name='retention_time',
            field=models.PositiveIntegerField(help_text='Retention time in days, if 0 - backup will be kept forever'),
        ),
    ]



from __future__ import unicode_literals

from waldur_mastermind.marketplace import views


def register_in(router):
    router.register(r'marketplace-service-providers', views.ServiceProviderViewSet,
                    base_name='marketplace-service-provider'),
    router.register(r'marketplace-categories', views.CategoryViewSet,
                    base_name='marketplace-category'),
    router.register(r'marketplace-offerings', views.OfferingViewSet,
                    base_name='marketplace-offering')
    router.register(r'marketplace-plans', views.PlanViewSet,
                    base_name='marketplace-plan')
    router.register(r'marketplace-screenshots', views.ScreenshotViewSet,
                    base_name='marketplace-screenshot')
    router.register(r'marketplace-orders', views.OrderViewSet,
                    base_name='marketplace-order'),



# -*- coding: utf-8 -*-
# Generated by Django 1.11.7 on 2018-07-03 17:43
from __future__ import unicode_literals

from django.db import migrations
import waldur_core.core.fields


class Migration(migrations.Migration):

    dependencies = [
        ('waldur_jira', '0018_project_runtime_state'),
    ]

    operations = [
        migrations.AlterField(
            model_name='project',
            name='action_details',
            field=waldur_core.core.fields.JSONField(default=dict),
        ),
    ]



from __future__ import unicode_literals

import collections
import cStringIO
import csv

from django.conf import settings
from django.contrib.contenttypes.models import ContentType
import python_freeipa

from waldur_core.structure import models as structure_models
from waldur_core.quotas import models as quota_models

from . import models, utils


class GroupSynchronizer(object):
    """
    This class maps Waldur structure units to FreeIPA groups and memberships.

    1) Customers and projects are modelled as FreeIPA groups.
    2) Group name for customer looks like waldur_org_590a409656054fc284b1102829406c63
       Group name for project looks like waldur_project_590a409656054fc284b1102829406c63
    2) Customer name is stored as group description.
       Similarly, project name is stored as group description.
    3) Project group is modelled as member of customer group.
    4) Permissions are mapped to group memberships.

    Synchronization is performed in batch processing task and consists of the following steps:

    1) Convert Waldur entities to group mappings.
    2) Convert FreeIPA entities to group mappings.
    3) Compare group mappings and apply series of API calls to synchronize them.

    Note that in order to distinguish between Waldur managed groups and internal FreeIPA groups,
    group name prefix should be specified. Similarly, there's also user name prefix setting available.
    """

    def __init__(self, client):
        self.client = client
        self.prefix = settings.WALDUR_FREEIPA['GROUPNAME_PREFIX']

        self.profiles = {
            profile.user_id: profile.username
            for profile in models.Profile.objects.all()
        }

        self.groups = set()
        self.group_users = collections.defaultdict(set)
        self.group_children = collections.defaultdict(set)
        self.group_names = dict()

        self.freeipa_groups = set()
        self.freeipa_users = collections.defaultdict(set)
        self.freeipa_children = collections.defaultdict(set)
        self.freeipa_names = dict()

    def group_name(self, key):
        return '%s%s' % (self.prefix, key)

    def project_group_name(self, project):
        return self.group_name('project_%s' % project.uuid)

    def customer_group_name(self, customer):
        return self.group_name('org_%s' % customer.uuid)

    def get_group_description(self, name, limit):
        stream = cStringIO.StringIO()
        writer = csv.writer(stream)
        writer.writerow([name.encode('utf-8'), str(limit)])
        return stream.getvalue().strip().decode('utf-8')

    def add_customer(self, customer, limit):
        group = self.customer_group_name(customer)
        self.groups.add(group)
        self.group_names[group] = self.get_group_description(customer.name, limit)

    def add_project(self, project, limit):
        project_group = self.project_group_name(project)
        self.groups.add(project_group)
        self.group_names[project_group] = self.get_group_description(project.name, limit)

        customer_group = self.customer_group_name(project.customer)
        self.groups.add(customer_group)

        self.group_children[customer_group].add(project_group)

    def add_customer_user(self, customer, user):
        username = self.profiles.get(user.id)
        if username:
            group = self.customer_group_name(customer)
            self.group_users[group].add(username)

    def add_project_user(self, project, user):
        username = self.profiles.get(user.id)
        if username:
            group = self.project_group_name(project)
            self.group_users[group].add(username)

    def collect_waldur_permissions(self):
        for permission in structure_models.CustomerPermission.objects.filter(is_active=True):
            self.add_customer_user(permission.customer, permission.user)

        for permission in structure_models.ProjectPermission.objects.filter(is_active=True):
            self.add_project_user(permission.project, permission.user)

    def get_limits(self, model):
        ctype = ContentType.objects.get_for_model(model)
        customer_quotas = quota_models.Quota.objects.filter(
            content_type=ctype, name=utils.QUOTA_NAME
        ).only('object_id', 'limit')
        return {
            quota.object_id: quota.limit
            for quota in customer_quotas
        }

    def collect_waldur_customers(self):
        limits = self.get_limits(structure_models.Customer)
        for customer in structure_models.Customer.objects.all():
            limit = limits.get(customer.id, -1.0)
            self.add_customer(customer, limit)

    def collect_waldur_projects(self):
        limits = self.get_limits(structure_models.Project)
        for project in structure_models.Project.objects.all():
            limit = limits.get(project.id, -1.0)
            self.add_project(project, limit)

    def add_freeipa_group(self, groupname, description, children):
        self.freeipa_groups.add(groupname)
        if description:
            self.freeipa_names[groupname] = description[0]
        self.freeipa_children[groupname] = set(child for child in children if child.startswith(self.prefix))

    def add_freeipa_users(self, groupname, users):
        self.freeipa_users[groupname].update(users)

    def collect_freeipa_groups(self):
        backend_groups = self.client.group_find()['result']
        for group in backend_groups:
            groupname = group['cn'][0]

            # Ignore groups not marked by own prefix
            if not groupname.startswith(self.prefix):
                continue

            members = group.get('member_user', [])
            description = group.get('description')
            children = group.get('member_group', [])
            self.add_freeipa_group(groupname, description, children)
            self.add_freeipa_users(groupname, members)

    def add_missing_groups(self):
        missing_groups = self.groups - self.freeipa_groups
        for group in missing_groups:
            utils.renew_task_status()
            self.client.group_add(group, description=self.group_names.get(group))

    def sync_group_names(self):
        for group in self.groups & self.freeipa_groups:
            utils.renew_task_status()
            waldur_name = self.group_names.get(group)
            freeipa_name = self.freeipa_names.get(group)
            if waldur_name != freeipa_name:
                self.client.group_mod(group, description=waldur_name)

    def sync_members(self):
        for group in self.groups:
            utils.renew_task_status()
            waldur_members = self.group_users.get(group, set())
            backend_members = self.freeipa_users.get(group, set())

            new_members = list(waldur_members - backend_members)
            if new_members:
                self.client.group_add_member(group, users=new_members, skip_errors=True)

            stale_members = list(backend_members - waldur_members)
            if stale_members:
                self.client.group_remove_member(group, users=stale_members, skip_errors=True)

    def sync_children(self):
        for group in self.groups:
            utils.renew_task_status()
            waldur_children = self.group_children.get(group, set())
            freeipa_children = self.freeipa_children.get(group, set())

            missing_children = list(waldur_children - freeipa_children)
            if missing_children:
                self.client.group_add_member(group, groups=missing_children, skip_errors=True)

            stale_children = list(freeipa_children - waldur_children)
            if stale_children:
                self.client.group_remove_member(group, groups=stale_children, skip_errors=True)

    def delete_stale_groups(self):
        for group in self.freeipa_groups - self.groups:
            utils.renew_task_status()
            self.client.group_del(group)

    def sync(self):
        try:
            self.collect_waldur_permissions()
            self.collect_waldur_customers()
            self.collect_waldur_projects()
            self.collect_freeipa_groups()

            self.add_missing_groups()
            self.sync_group_names()
            self.sync_members()
            self.sync_children()
            self.delete_stale_groups()

        finally:
            utils.release_task_status()


class FreeIPABackend(object):
    def __init__(self):
        options = settings.WALDUR_FREEIPA
        self._client = python_freeipa.Client(
            host=options['HOSTNAME'],
            verify_ssl=options['VERIFY_SSL']
        )
        self._client.login(options['USERNAME'], options['PASSWORD'])

    def _format_ssh_keys(self, user):
        return list(user.sshpublickey_set.values_list('public_key', flat=True))

    def create_profile(self, profile):
        waldur_user = profile.user
        ssh_keys = self._format_ssh_keys(waldur_user)
        first_name, last_name, _ = utils.get_names(profile.user.full_name)

        self._client.user_add(
            username=profile.username,
            first_name=first_name,
            last_name=last_name,
            full_name=waldur_user.full_name,
            mail=waldur_user.email,
            job_title=waldur_user.job_title,
            preferred_language=waldur_user.preferred_language,
            telephonenumber=waldur_user.phone_number,
            ssh_key=ssh_keys,
            gecos=profile.gecos,
        )

    def disable_profile(self, profile):
        self._client.user_disable(profile.username)

    def enable_profile(self, profile):
        self._client.user_enable(profile.username)

    def update_ssh_keys(self, profile):
        ssh_keys = self._format_ssh_keys(profile.user)
        if ssh_keys:
            ssh_keys = sorted(ssh_keys)
        else:
            ssh_keys = None

        backend_profile = self._client.user_show(profile.username)
        backend_keys = backend_profile.get('ipasshpubkey')
        if backend_keys:
            backend_keys = sorted(backend_keys)

        if backend_keys != ssh_keys:
            self._client.user_mod(profile.username, ipasshpubkey=ssh_keys)

    def _update_profile(self, profile, params):
        try:
            self._client.user_mod(profile.username, **params)
        except python_freeipa.exceptions.BadRequest as e:
            # If no modifications to be performed freeipa-server return an exception.
            if e.code == 4202:
                pass

    def update_name(self, profile):
        first_name, last_name, initials = utils.get_names(profile.user.full_name)
        params = {
            'givenname': first_name,
            'sn': last_name,
            'cn': profile.user.full_name,
            'displayname': profile.user.full_name,
            'initials': initials,
        }
        self._update_profile(profile, params)

    def update_gecos(self, profile):
        params = {
            'gecos': profile.gecos,
        }
        self._update_profile(profile, params)

    def synchronize_names(self):
        for profile in models.Profile.objects.filter(is_active=True):
            self.update_name(profile)

    def synchronize_gecos(self):
        for profile in models.Profile.objects.filter(is_active=True):
            self.update_gecos(profile)

    def synchronize_groups(self):
        synchronizer = GroupSynchronizer(self._client)
        synchronizer.sync()



from waldur_core.logging.loggers import EventLogger, event_logger


class InvoiceLogger(EventLogger):
    month = int
    year = int
    customer = 'structure.Customer'

    class Meta:
        event_types = ('invoice_created', 'invoice_paid', 'invoice_canceled')
        event_groups = {
            'customers': event_types,
            'invoices': event_types,
        }


event_logger.register('invoice', InvoiceLogger)



from __future__ import unicode_literals

from datetime import timedelta

from waldur_core.core import WaldurExtension


class AnalyticsExtension(WaldurExtension):
    class Settings:
        # See also: http://influxdb-python.readthedocs.io/en/latest/api-documentation.html#influxdbclient
        WALDUR_ANALYTICS = {
            'ENABLED': False,
            'INFLUXDB': {
                'host': 'localhost',
                'port': 8086,
                'username': 'USERNAME',
                'password': 'PASSWORD',
                'database': 'DATABASE',
                'ssl': False,
                'verify_ssl': False,
            }
        }

    @staticmethod
    def django_app():
        return 'waldur_mastermind.analytics'

    @staticmethod
    def is_assembly():
        return True

    @staticmethod
    def celery_tasks():
        return {
            'waldur-push-analytics': {
                'task': 'analytics.push_points',
                'schedule': timedelta(minutes=30),
                'args': (),
            },
        }



from waldur_core.core.managers import GenericKeyMixin
from waldur_core.structure.managers import StructureManager
from waldur_core.structure.models import NewResource as Resource, ResourceMixin


def filter_active(qs):
    INVALID_STATES = (
        Resource.States.CREATION_SCHEDULED,
        Resource.States.DELETION_SCHEDULED,
        Resource.States.DELETING,
        Resource.States.ERRED
    )
    return qs.exclude(backend_id='', state__in=INVALID_STATES)


class HostManager(GenericKeyMixin, StructureManager):
    """ Allows to filter and get hosts by generic key """

    def get_available_models(self):
        """ Return list of models that are acceptable """
        return ResourceMixin.get_all_models()



from __future__ import unicode_literals

from django.utils.translation import ugettext_lazy as _
from rest_framework import exceptions, response, status, views

from waldur_core.core import views as core_views
from waldur_core.structure import models as structure_models
from waldur_core.structure import filters as structure_filters
from waldur_mastermind.invoices import models as invoices_models
from waldur_mastermind.invoices import utils as invoice_utils

from . import filters, models, serializers


class PriceEstimateViewSet(core_views.ActionsViewSet):
    disabled_actions = ['create', 'destroy']
    queryset = models.PriceEstimate.objects.all()
    serializer_class = serializers.PriceEstimateSerializer
    lookup_field = 'uuid'
    filter_backends = (
        filters.PriceEstimateScopeFilterBackend,
    )

    def get_queryset(self):
        return models.PriceEstimate.objects.filtered_for_user(self.request.user)

    def is_owner_or_staff(request, view, obj=None):
        if not obj:
            return False

        if request.user.is_staff:
            return True

        if isinstance(obj.scope, structure_models.Customer):
            raise exceptions.PermissionDenied(
                _('Only staff is allowed to modify policy for the customer.')
            )

        elif isinstance(obj.scope, structure_models.Project):
            customer = obj.scope.customer
            if not customer.has_user(request.user, structure_models.CustomerRole.OWNER):
                raise exceptions.PermissionDenied(
                    _('Only staff and customer owner is allowed to modify policy for the project.')
                )

    update_permissions = [is_owner_or_staff]


class TotalCustomerCostView(views.APIView):
    def get(self, request, format=None):
        if not self.request.user.is_staff and not request.user.is_support:
            raise exceptions.PermissionDenied()

        customers = structure_models.Customer.objects.all()
        customers = structure_filters.AccountingStartDateFilter().filter_queryset(request, customers, self)

        name = request.query_params.get('name', '')
        if name:
            customers = customers.filter(name__icontains=name)

        year, month = invoice_utils.parse_period(request.query_params)
        invoices = invoices_models.Invoice.objects.filter(customer__in=customers)
        invoices = invoices.filter(year=year, month=month)

        total = sum(invoice.total for invoice in invoices)
        return response.Response({'total': total}, status=status.HTTP_200_OK)
